#include"../sys.h"

#include<stdint.h>
#include<stddef.h>
#include"../std.h"
#include"fs.h"

asm("modentry:\n"
"mov $stack + 2048, %esp\n"
"call startc");

char stack[3072];

// Should be renamed. This *isn't* LeanFS, however inspired it may be.

#define EXTENTS_PER_INODE 6

#define INODE_ftRegular 1
#define INODE_ftDirectory 2
#define INODE_ftMask 7

#define INODE_ATTRIBUTE_iaInlineExtAttr (1 << 19)
#define INODE_ATTRIBUTE_iaFmtRegular (1 << 29)
#define INODE_ATTRIBUTE_iaFmtDirectory (2 << 29)
#define INODE_ATTRIBUTE_iaFmtMask (7 << 29)


struct Superblock {
	uint32_t checksum;
	uint32_t magic;
	uint16_t fsVersion;
	uint8_t preallocCount;
	uint8_t logBlocksPerBand;
	uint32_t state;
	uint8_t uuid[16];
	char volumeLabel[64];
	uint64_t blockCount;
	uint64_t freeBlockCount;
	uint64_t primarySuper;
	uint64_t backupSuper;
	uint64_t bitmapStart;
	uint64_t rootInode;
	uint64_t badInode;
	uint64_t journalInode;
	uint8_t logBlockSize;
	uint8_t reserved0[7];
} __attribute__((packed, aligned(4)));

struct INode {
	uint32_t checksum;
	uint32_t magic;
	uint8_t extentCount;
	uint8_t reserved[3];
	uint32_t indirectCount;
	uint32_t linkCount;
	uint32_t uid;
	uint32_t gid;
	uint32_t attributes;
	uint64_t fileSize;
	uint64_t blockCount;
	int64_t accessTime;
	int64_t statusChangeTime;
	int64_t modificationTime;
	int64_t creationTime;
	uint64_t firstIndirect;
	uint64_t lastIndirect;
	uint64_t fork;
	uint64_t extentStarts[EXTENTS_PER_INODE];
	uint32_t extentSizes[EXTENTS_PER_INODE];
} __attribute__((packed));

struct DirectoryEntry {
	uint64_t inode;
	uint8_t type;
	uint8_t recLen;
	uint16_t nameLen;
	uint8_t name[];
} __attribute__((packed));

uint8_t *diskWindow = (uint8_t*) -1;

size_t volumeStartLBA, superblockLBA;
struct Superblock superblock;

void read(size_t start, size_t length) {
	((uint64_t*) diskWindow)[0] = start;
	((uint32_t*) diskWindow)[2] = length;
	
	memory_barrier();
	sys_signal_send(diskWindow, 0);
	
	sys_signal_wait(diskWindow, NULL);
}

static uint32_t lean_compute_checksum(const uint32_t *buf, size_t dwordLength, size_t dwordPadTo) {
	uint32_t sum = 0;
	
	size_t i = 1;
	for(; i < dwordLength; i++) {
		sum = (sum << 31) + (sum >> 1);
		sum += buf[i];
	}
	
	for(; i < dwordPadTo; i++) {
		sum = (sum << 31) + (sum >> 1);
	}
	
	return sum;
}

int parse_inode(size_t id) {
	read(volumeStartLBA * 512 + (id << superblock.logBlockSize), 1 << superblock.logBlockSize);
	
	struct INode *test = (struct INode*) diskWindow;
	
	if(test->magic != 0x45444F4E) {
		return 0;
	}
	
	// Driver has outdated checksums
#if 0
	if(lean_compute_checksum((uint32_t*) test, (sizeof(struct INode)) >> 2, (sizeof(struct INode)) >> 2) != test->checksum) {
		return 0;
	}
#endif
	
	return 1;
}

size_t lean_read_inode(const struct INode *inode, size_t off, size_t len, uint8_t *ret) {
	if(off + len > inode->fileSize) {
		len = inode->fileSize - off;
	}
	
	if(inode->attributes & INODE_ATTRIBUTE_iaInlineExtAttr) {
		off += 1 << superblock.logBlockSize;
	} else {
		off += sizeof(struct INode);
	}
	
	size_t startBlk = off >> superblock.logBlockSize;
	size_t endBlk = (off + len - 1) >> superblock.logBlockSize;
	
	uint32_t b = 0;
	uint32_t ei, eb;
	for(ei = 0; ei < EXTENTS_PER_INODE; ei++) {
		for(eb = 0; eb < inode->extentSizes[ei]; eb++) {
			if(b == startBlk) {
				goto foundStart;
			}
			
			b++;
		}
	}
	
	// Not found. Corruption?
	return 0;
	
foundStart:
	
	while(1) {
		read(volumeStartLBA * 512 + ((inode->extentStarts[ei] + eb) << superblock.logBlockSize), 1 << superblock.logBlockSize);
		
		if(b == startBlk && b == endBlk) {
			std_copy(ret, diskWindow + off, len);
		
			break;
		} else if(b == startBlk) {
			uint16_t mod = off % (1 << superblock.logBlockSize);
			
			std_copy(ret, diskWindow + mod, (1 << superblock.logBlockSize) - mod);
			
			ret += (1 << superblock.logBlockSize) - mod;
		} else if(b == endBlk) {
			uint16_t mod = (off + len - 1) % (1 << superblock.logBlockSize);
			
			std_copy(ret, diskWindow, mod + 1);
			
			break;
		} else {
			std_copy(ret, diskWindow, 1 << superblock.logBlockSize);
			
			ret += 1 << superblock.logBlockSize;
		}
		
		eb++, b++;
		
		if(eb == inode->extentSizes[ei]) {
			eb = 0;
			ei++;
			
			if(ei == EXTENTS_PER_INODE) {
				// Stop. Indirect blocks not yet supported.
				return 0;
			}
		}
	}
	
	return len;
}

// I don't like this.
// Biggest ugly by far is how it operates in 512-byte chunks, even though lean_read_inode already handles unaligned access
uint64_t find_child(const struct INode *node, const Str16 *name) {
	uint8_t buf[512] __attribute__((aligned(16)));
	
	size_t offset = 0;
	lean_read_inode(node, offset, sizeof(buf), buf);
	
	struct DirectoryEntry *e = (struct DirectoryEntry*) buf;
	while(1) {
		if(offset + ((uintptr_t) e - (uintptr_t) buf) >= node->fileSize) {
			// Reached EOF.
			return 0;
		} else while((uintptr_t) e >= (uintptr_t) buf + sizeof(buf)) {
			offset += sizeof(buf);
			lean_read_inode(node, offset, sizeof(buf), buf);
			e = (struct DirectoryEntry*) (((uintptr_t) e - (uintptr_t) buf) % sizeof(buf) + (uintptr_t) buf);
		}
		
		if(e->nameLen != name->len) {
			uintptr_t after = (uintptr_t) e + 16 * e->recLen;
			
			e = (struct DirectoryEntry*) after;
			
			goto next;
		}
		
		// Must be saved, as `e` becomes invalid in the following for loop
		uint64_t childinode = e->inode;
		size_t nameLen = e->nameLen;
		uint8_t *testname = e->name;
		for(uint16_t i = 0; i < nameLen; i++) {
			if(testname - buf == sizeof(buf)) {
				offset += sizeof(buf);
				lean_read_inode(node, offset, sizeof(buf), buf);
				testname = buf;
			}
			if(*testname != name->data[i]) {
				testname += nameLen - i;
				
				e = (struct DirectoryEntry*) ((((uintptr_t) testname) + 15) & ~15);
				
				goto next;
			}
			testname++;
		}
		
		return childinode;

next:
	}
}

uint64_t get_inode_from_str16_path(const Str16 *path, struct INode *data) {
	uint64_t child = superblock.rootInode;
	if(!parse_inode(child)) {
		return 0;
	}
	std_copy(data, diskWindow, sizeof(*data));
	
	while(path->len) {
		child = find_child(data, path);
		
		if(!child) {
			return 0;
		}
		
		if(!parse_inode(child)) {
			return 0;
		}
		std_copy(data, diskWindow, sizeof(*data));
		
		path = (const Str16*) ((uintptr_t) path + sizeof(*path) + path->len);
	}
	
	return child;
}

void startc() {
	do {
		diskWindow = sys_link_create('STRG', 'UD\0\0', 4096);
	} while(diskWindow == (void*) -1);
	
	read(0, 512);
	volumeStartLBA = *(uint32_t*) &diskWindow[0x1C6];
	
	int found = 0;
	for(size_t superblockLBA = 0; superblockLBA < 256; superblockLBA++) {
		read((volumeStartLBA + superblockLBA) * 512, 512);
		
		struct Superblock *candidate = (struct Superblock*) diskWindow;
		
		if(candidate->magic == 0x4E41454C && lean_compute_checksum((uint32_t*) diskWindow, sizeof(struct Superblock) >> 2, (1 << candidate->logBlockSize) >> 2) == candidate->checksum) {
			superblock = *(struct Superblock*) diskWindow;
			found = 1;
			break;
		}
	}
	
	if(found) {
		uint32_t canalID = sys_canal_create('FS\0\0', 'ROOT');
		while(1) {
			sys_canal_accept(canalID);
			
			uint8_t *fsBuf = sys_signal_wait(SYS_SIGNAL_WAIT_ANY, NULL);
			memory_barrier();
			
			LumaFSCommand *cmd = (void*) fsBuf;
			if(cmd->op == LUMA_FS_COMMAND_READ) {
				struct INode data;
				uint64_t inode = get_inode_from_str16_path((Str16*) cmd->read.path, &data);
				
				if(inode) {
					size_t offset = cmd->read.offset;
					uint32_t left = cmd->read.bytes;
					
					if(left) {
						while(1) {
							size_t amount = left > 4096 ? 4096 : left;
							
							size_t actuallyRead = lean_read_inode(&data, offset, amount, fsBuf);
							if(actuallyRead < amount) {
								amount = actuallyRead;
								left = actuallyRead;
							}
							
							memory_barrier();
							sys_signal_send(fsBuf, amount);
							
							if(left == 4096) {
								sys_signal_wait(fsBuf, NULL);
								sys_signal_send(fsBuf, 0);
								break;
							} else {
								offset += amount;
								left -= amount;
								
								if(left == 0) {
									break;
								} else {
									sys_signal_wait(fsBuf, NULL);
								}
							}
						}
					}
				} else {
					memory_barrier();
					sys_signal_send(fsBuf, 0);
				}
			} else if(cmd->op == LUMA_FS_COMMAND_INFO) {
				struct INode data;
				uint64_t inode = get_inode_from_str16_path((Str16*) cmd->info.path, &data);
				
				prcs("Requesting info ");
				prs(cmd->info.path);
				pr('\n');
				
				prcs("Got inode ");
				pri(inode);
				pr('\n');
				
				if(inode) {
					prcs("Size ");
					pri(data.fileSize);
					pr('\n');
					
					prcs("magic ");
					pri(data.magic);
					pr('\n');
					
					cmd->info.fileSize = data.fileSize;
					
					memory_barrier();
					sys_signal_send(fsBuf, 1);
				} else {
					memory_barrier();
					sys_signal_send(fsBuf, 0);
				}
			} else if(cmd->op == LUMA_FS_COMMAND_TRAVERSE) {
				struct INode data;
				uint64_t inode = get_inode_from_str16_path((Str16*) cmd->traverse.path, &data);
				
				if(inode) {
					size_t offset = 0, left = data.fileSize;
					uint8_t buf[512] __attribute__((aligned(16)));
					
					lean_read_inode(&data, offset, sizeof(buf), buf);
					
					struct DirectoryEntry *in = (struct DirectoryEntry*) buf;
					
					LumaFSDirectoryEntry *out = (LumaFSDirectoryEntry*) fsBuf;
					
					while(1) {
						size_t len = in->nameLen > 255 ? 255 : in->nameLen;
						
						if((uintptr_t) out + sizeof(*buf) + len > (uintptr_t) fsBuf + 4096) {
							memory_barrier();
							sys_signal_send(fsBuf, 2);
							sys_signal_wait(fsBuf, NULL);
						}
						
						out->type = ((in->type & INODE_ftMask) == INODE_ftDirectory) ? LUMA_FS_DIR_ENTRY_ISDIR : 0;
						out->name.len = len;
						uint8_t *nameIn = in->name;
						for(size_t n = 0; n < out->name.len; n++) {
							out->name.data[n] = *(nameIn++);
							if((uintptr_t) nameIn == (uintptr_t) buf + sizeof(buf)) {
								offset += sizeof(buf);
								left -= sizeof(buf);
								lean_read_inode(&data, offset, sizeof(buf), buf);
								nameIn = buf;
							}
						}
						
						in = (struct DirectoryEntry*) (((uintptr_t) nameIn + 15) & ~15);
						if(offset + ((uintptr_t) in - (uintptr_t) buf) == data.fileSize) {
							out->type |= LUMA_FS_DIR_ENTRY_LAST;
							break;
						} else if((uintptr_t) in == (uintptr_t) in + sizeof(buf)) {
							offset += sizeof(buf);
							left -= sizeof(buf);
							lean_read_inode(&data, offset, sizeof(buf), buf);
							in = (struct DirectoryEntry*) buf;
						}
						
						out = (LumaFSDirectoryEntry*) ((uintptr_t) out + sizeof(*out) + len);
					}
					
					memory_barrier();
					sys_signal_send(fsBuf, 1);
				} else {
					memory_barrier();
					sys_signal_send(fsBuf, 0);
				}
			} else {
				memory_barrier();
				sys_signal_send(fsBuf, 0);
			}
		}
	} else {
		while(1) { sys_signal_wait(SYS_SIGNAL_WAIT_ANY, NULL); }
	}
}
