#pragma once

#include<stdint.h>
#include<stdbool.h>
#include<stddef.h>
#include"tiny.h"

struct k3Font;
struct k3MObj;

#define k3M_EVENT_MOUSE_ENTER 0
#define k3M_EVENT_MOUSE_MOTION 1
#define k3M_EVENT_MOUSE_LEAVE 2
#define k3M_EVENT_MOUSE_PRESS 3
#define k3M_EVENT_MOUSE_RELEASE 4
#define k3M_EVENT_MOUSE_CLICK 5
#define k3M_EVENT_KEY_PRESS 6
#define k3M_EVENT_KEY_RELEASE 7
#define k3M_EVENT_INPUT 8
#define k3M_EVENT_DRAW 9
#define k3M_EVENT_COMPLETE 10
#define k3M_EVENT_ALL 11

#define k3M_MOUSE_BUTTON_0 0
#define k3M_MOUSE_BUTTON_1 1
#define k3M_MOUSE_BUTTON_2 2

#define k3M_EVENTKIND_DIRECT 0
#define k3M_EVENTKIND_CAPTURE 1
#define k3M_EVENTKIND_BUBBLE 2

#define k3M_USERDATA_SIZE 16

#include"../std.h"

struct k3MEvent {
	uint16_t code;
	
	uint8_t kind;
	
	struct k3MObj *original;
	struct k3MObj *target;
	
	union {
		struct {
			int button;
			int16_t x;
			int16_t y;
		} mouse;
		struct {
			uint32_t num;
		} key;
		struct {
			uint32_t code;
		} input;
	};
};

typedef bool(*k3MEventHandlerFunc)(struct k3MEvent*, uint8_t *ud);

typedef struct k3MEventHandler {
	uint8_t ud[k3M_USERDATA_SIZE];
	k3MEventHandlerFunc callback;
	uint16_t code;
} k3MEventHandler;

enum k3MPropertyType {
	k3M_PROP_BG_COLOR,
	k3M_PROP_BORDER_RADIUS,
	k3M_PROP_MARGIN,
};
struct k3MProperty {
	enum k3MPropertyType type;
	union {
		intmax_t si[4];
		void *ptr;
		uint8_t buf[32];
	};
};

typedef void(*k3MLayout)(struct k3MObj*);

struct k3MObj {
	struct k3MObj *parent;
	
	k3MLayout layout;
	
	size_t childCount;
	struct k3MObj **children;
	
	int16_t x;
	int16_t y;
	int16_t w;
	int16_t h;
	
	bool invisible, hovered, disabled;
	
	k3MEventHandler *handlers;
	size_t handlerCount;
	
	struct k3MProperty *properties;
	size_t propertyCount;
};
struct k3MObj *k3MObj();

#define k3MenuSetBounds(a, x, y, w, h) k3MenuSetBounds_((struct k3MObj*) (a), (x), (y), (w), (h))
void k3MenuSetBounds_(struct k3MObj *this, int16_t x, int16_t y, int16_t w, int16_t h);

#define k3MRegisterEventHandler(a, c, cb, ud, udsz) k3MRegisterEventHandler_((struct k3MObj*) (a), (c), (cb), (ud), (udsz))
int k3MRegisterEventHandler_(struct k3MObj *obj, uint16_t evcode, k3MEventHandlerFunc callback, uint8_t *ud, size_t udSize);

bool k3MEventSend(struct k3MEvent *ev);

int k3MRemoveChild(struct k3MObj *parent, struct k3MObj *child);
void k3MAddChild(struct k3MObj *parent, struct k3MObj *child);

void k3MOverrideProperty(struct k3MObj *obj, struct k3MProperty);
struct k3MProperty *k3MFindProperty(struct k3MObj *obj, enum k3MPropertyType, bool direct);

struct k3MLabel {
	struct k3MObj;
	
	struct k3Font *font;
	uint32_t sz;
	char *txt;
};
struct k3MLabel *k3MLabel(struct k3Font *font, uint32_t millipts, const char *txt);

struct k3MScreen {
	struct k3MObj;
	
	struct k3MObj *keyboardFocus;
	struct k3MObj *lastHover;
};
struct k3MScreen *k3MScreen();

struct k3MTextButton {
	struct k3MObj;
	
	struct k3Font *font;
	uint32_t sz;
	char *txt;
};
struct k3MTextButton *k3MTextButton(struct k3Font *font, uint32_t millipts, const char *txt);

struct k3MTextInput {
	struct k3MObj;
	
	struct k3Font *font;
	uint32_t sz;
	
	char *placeholder;
	
	char *txt;
};
struct k3MTextInput *k3MTextInput(struct k3Font *font, uint32_t millipts, const char *placeholder, const char *txt);

static inline int k3UTF8Encode(uint32_t cp, uint8_t ret[static 4]) {
	if(cp < 0x80) {
		ret[0] = cp;
		return 1;
	} else if(cp < 0x800) {
		ret[0] = 192 | (cp >> 6);
		ret[1] = 128 | (cp & 63);
		return 2;
	} else if(cp < 0x10000) {
		ret[0] = 224 | (cp >> 12);
		ret[1] = 128 | ((cp >> 6) & 63);
		ret[2] = 128 | (cp & 63);
		return 3;
	} else if(cp < 0x110000) {
		ret[0] = 240 | (cp >> 18);
		ret[1] = 128 | ((cp >> 12) & 63);
		ret[2] = 128 | ((cp >> 6) & 63);
		ret[3] = 128 | (cp & 63);
		return 4;
	}
	return 0;
}

static inline size_t strlen(const char *str) {
	size_t sz = 0;
	while(*str) {
		sz++;
		str++;
	}
	return sz;
}

static inline const char *k3UTF8LastCodepointZ(const char *txt) {
	size_t len = strlen(txt);
	while(len--) {
		if((txt[len] & 0xC0) != 0x80) {
			return &txt[len];
		}
	}
	return NULL;
}

static inline char *strdup(const char *s) {
	size_t len = strlen(s);
	char *d = tiny_malloc(len + 1);
	d[len] = '\0';
	__builtin_memcpy(d, s, len);
	return d;
}

